package systems.kumi.simpleblereceiver

import android.bluetooth.BluetoothAdapter
import android.bluetooth.BluetoothDevice
import android.bluetooth.BluetoothGatt
import android.bluetooth.BluetoothGattCallback
import android.bluetooth.BluetoothGattCharacteristic
import android.bluetooth.BluetoothManager
import android.bluetooth.le.ScanCallback
import android.bluetooth.le.ScanResult
import android.Manifest
import android.annotation.SuppressLint
import android.content.BroadcastReceiver
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.content.pm.PackageManager
import android.os.Bundle
import android.os.ParcelUuid
import androidx.activity.ComponentActivity
import androidx.activity.compose.setContent
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.material3.Button
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.Surface
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.ui.Modifier
import androidx.compose.ui.tooling.preview.Preview
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat
import systems.kumi.simpleblereceiver.ui.theme.SimpleBLEReceiverTheme
import java.util.UUID

class MainActivity : ComponentActivity() {
    private lateinit var bluetoothAdapter: BluetoothAdapter
    private var bluetoothGatt: BluetoothGatt? = null

    companion object {
        val SERVICE_UUID = UUID.fromString("14080001-1337-BABE-B00B-C0FFEEC0FFEE")
        val CHARACTERISTIC_UUID = UUID.fromString("14080002-1337-BABE-B00B-C0FFEEC0FFEE")
        const val REQUEST_ENABLE_BT = 1
        const val PERMISSION_REQUEST_CODE = 101
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        val bluetoothManager = getSystemService(Context.BLUETOOTH_SERVICE) as BluetoothManager
        bluetoothAdapter = bluetoothManager.adapter

        // Check for permissions and Bluetooth availability
        if (ContextCompat.checkSelfPermission(this, Manifest.permission.ACCESS_FINE_LOCATION)
            != PackageManager.PERMISSION_GRANTED
        ) {
            ActivityCompat.requestPermissions(
                this,
                arrayOf(Manifest.permission.ACCESS_FINE_LOCATION),
                PERMISSION_REQUEST_CODE
            )
        } else if (bluetoothAdapter.isEnabled) {
            startScanning()
        } else {
            val enableBtIntent = Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE)
            startActivityForResult(enableBtIntent, REQUEST_ENABLE_BT)
        }

        val intentFilter = IntentFilter("systems.kumi.simplerfiddemo.NEW_TAG")
        registerReceiver(appReceiver, intentFilter)

        setContent {
            SimpleBLEReceiverTheme {
                Surface(
                    modifier = Modifier.fillMaxSize(),
                    color = MaterialTheme.colorScheme.background
                ) {
                    Column {
                        Text("Waiting for data...")

                        Button(onClick = { emitIntent("sample-id") }) {
                            Text("Emit sample intent")
                        }
                    }
                }
            }
        }
    }

    private val appReceiver = object : BroadcastReceiver() {
        override fun onReceive(context: Context?, intent: Intent?) {
            val data = intent?.getStringExtra("data")
            if (data != null) {
                runOnUiThread {
                    setContent {
                        SimpleBLEReceiverTheme {
                            Surface(
                                modifier = Modifier.fillMaxSize(),
                                color = MaterialTheme.colorScheme.background
                            ) {
                                Column {
                                    Text("Received data: $data (from intent)")

                                    Button(onClick = { emitIntent("sample-id") }) {
                                        Text("Emit sample data")
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    private fun startScanning() {
        val leScanner = bluetoothAdapter.bluetoothLeScanner
        val settings = android.bluetooth.le.ScanSettings.Builder()
            .setScanMode(android.bluetooth.le.ScanSettings.SCAN_MODE_LOW_LATENCY)
            .build()

        val filter = android.bluetooth.le.ScanFilter.Builder()
            .setServiceUuid(ParcelUuid(SERVICE_UUID))
            .build()

        if (ActivityCompat.checkSelfPermission(
                this,
                android.Manifest.permission.BLUETOOTH_SCAN
            ) != PackageManager.PERMISSION_GRANTED
        ) {
            ActivityCompat.requestPermissions(
                this,
                arrayOf(
                    android.Manifest.permission.BLUETOOTH_SCAN,
                    android.Manifest.permission.BLUETOOTH_ADMIN,
                    android.Manifest.permission.BLUETOOTH,
                    android.Manifest.permission.BLUETOOTH_CONNECT,
                    android.Manifest.permission.BLUETOOTH_ADVERTISE
                ),
                1
            )
            return
        }
        leScanner.startScan(listOf(filter), settings, scanCallback)
    }

    private val scanCallback = object : ScanCallback() {
        override fun onScanResult(callbackType: Int, result: ScanResult?) {
            super.onScanResult(callbackType, result)
            if (result == null || result.device == null) return

            connectToDevice(result.device)
        }

        override fun onBatchScanResults(results: MutableList<ScanResult>?) {
            super.onBatchScanResults(results)
            if (results.isNullOrEmpty()) return

            results.forEach { result ->
                connectToDevice(result.device)
            }
        }
    }

    private fun connectToDevice(device: BluetoothDevice) {
        if (ActivityCompat.checkSelfPermission(
                this,
                android.Manifest.permission.BLUETOOTH_CONNECT
            ) != PackageManager.PERMISSION_GRANTED
        ) {
            ActivityCompat.requestPermissions(
                this,
                arrayOf(
                    android.Manifest.permission.BLUETOOTH_CONNECT,
                    android.Manifest.permission.BLUETOOTH_ADMIN,
                    android.Manifest.permission.BLUETOOTH,
                    android.Manifest.permission.BLUETOOTH_SCAN,
                    android.Manifest.permission.BLUETOOTH_ADVERTISE
                ),
                1
            )
            return
        }
        bluetoothGatt = device.connectGatt(this, false, gattCallback)
    }

    private fun emitIntent(data: String) {
        val intent = Intent("systems.kumi.simpleblereceiver.DATA_RECEIVED")
        intent.putExtra("data", data)
        sendBroadcast(intent)
    }

    private val gattCallback = object : BluetoothGattCallback() {
        @SuppressLint("MissingPermission")
        override fun onConnectionStateChange(gatt: BluetoothGatt?, status: Int, newState: Int) {
            super.onConnectionStateChange(gatt, status, newState)
            if (newState == BluetoothGatt.STATE_CONNECTED) {
                gatt?.discoverServices()
            } else if (newState == BluetoothGatt.STATE_DISCONNECTED) {
                bluetoothGatt = null
            }
        }

        @SuppressLint("MissingPermission")
        override fun onServicesDiscovered(gatt: BluetoothGatt?, status: Int) {
            super.onServicesDiscovered(gatt, status)
            if (status == BluetoothGatt.GATT_SUCCESS) {
                gatt?.services?.forEach { service ->
                    if (service.uuid == SERVICE_UUID) {
                        val characteristic = service.getCharacteristic(CHARACTERISTIC_UUID)
                        gatt.readCharacteristic(characteristic)
                    }
                }
            }
        }

        override fun onCharacteristicRead(
            gatt: BluetoothGatt?,
            characteristic: BluetoothGattCharacteristic?,
            status: Int
        ) {
            super.onCharacteristicRead(gatt, characteristic, status)
            if (status == BluetoothGatt.GATT_SUCCESS) {
                if (characteristic?.uuid == CHARACTERISTIC_UUID) {
                    val data = characteristic!!.value
                    val dataString = data?.joinToString(separator = "") { byte -> "%02X".format(byte) }

                    runOnUiThread {
                        setContent {
                            SimpleBLEReceiverTheme {
                                Surface(
                                    modifier = Modifier.fillMaxSize(),
                                    color = MaterialTheme.colorScheme.background
                                ) {
                                    Column {
                                        Text("Connected to device: ${gatt?.device?.address}")
                                        Text("Received data: $dataString")

                                        Button(onClick = { emitIntent("sample-id") }) {
                                            Text("Emit sample data")
                                        }
                                    }
                                }
                            }
                        }
                    }

                    if (dataString != null) {
                        emitIntent(dataString)
                    }
                }
            }
        }
    }
}

@Composable
fun ReceiverStatus(deviceAddress: String? = null, data: String? = null) {
    Column {
        Text(text = "Device Address: ${deviceAddress ?: "Not connected"}")
        Text(text = "Data Received: ${data ?: "No data"}")

        Button(onClick = {}) {
            Text(text = "Emit sample data")
        }
    }
}

@Preview(showBackground = true)
@Composable
fun DefaultPreview() {
    SimpleBLEReceiverTheme {
        ReceiverStatus()
    }
}